import json
import logging
import sys
from typing import Optional

from ..authclient import AuthClient
from ..common.schemas import StateRead
from ..common.schemas import TaskCollectPip
from ..common.schemas import TaskCreate
from ..common.schemas import TaskRead
from ..common.schemas import TaskUpdate
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichJsonInterface
from ..response import check_response
from ._aux_task_caching import FractalCacheError
from ._aux_task_caching import get_task_id_from_cache
from ._aux_task_caching import refresh_task_cache


async def get_task_list(client: AuthClient) -> RichJsonInterface:
    task_list = await refresh_task_cache(client=client)
    return RichJsonInterface(retcode=0, data=task_list)


async def task_collect_pip(
    client: AuthClient,
    *,
    package: str,
    package_version: Optional[str] = None,
    python_version: Optional[str] = None,
    package_extras: Optional[str] = None,
    pinned_dependency: Optional[list[str]] = None,
    batch: bool = False,
) -> BaseInterface:

    # Construct TaskCollectPip object
    attributes = dict(package=package)
    if package_version:
        attributes["package_version"] = package_version
    if python_version:
        attributes["python_version"] = python_version
    if package_extras:
        attributes["package_extras"] = package_extras
    if pinned_dependency:
        for pin in pinned_dependency:
            if len(pin.split("=")) != 2:
                logging.error(
                    f"Invalid pin: {pin}.\nPins must be written as "
                    "'--pinned-dependency PACKAGE_NAME=PACKAGE_VERSION'"
                )
                sys.exit(1)
        attributes["pinned_package_versions"] = {
            _name: _version
            for _name, _version in (p.split("=") for p in pinned_dependency)
        }
    task_collect = TaskCollectPip(**attributes)

    res = await client.post(
        f"{settings.BASE_URL}/task/collect/pip/",
        json=task_collect.dict(exclude_unset=True),
    )

    state = check_response(
        res, expected_status_code=[200, 201], coerce=StateRead
    )
    if batch:
        output = f"{state.id} {state.data['venv_path']}"
        return PrintInterface(retcode=0, data=output)
    else:
        return RichJsonInterface(retcode=0, data=state.sanitised_dict())


async def task_collection_check(
    client: AuthClient,
    *,
    state_id: int,
    include_logs: bool,
    do_not_separate_logs: bool = False,
) -> BaseInterface:

    res = await client.get(
        f"{settings.BASE_URL}/task/collect/{state_id}?verbose={include_logs}"
    )
    state = check_response(res, expected_status_code=200, coerce=StateRead)

    state_dict = state.sanitised_dict()

    # Remove key-value pairs with None value
    state_dict["data"] = {
        key: val for (key, val) in state_dict["data"].items() if val
    }

    if (not include_logs) or do_not_separate_logs:
        return RichJsonInterface(retcode=0, data=state_dict)
    else:
        log = state_dict["data"].pop("log")
        extra_lines = ["\nThis is the task-collection log:\n", log]
        return RichJsonInterface(
            retcode=0, data=state_dict, extra_lines=extra_lines
        )


async def post_task(
    client: AuthClient,
    *,
    name: str,
    command: str,
    source: str,
    input_type: str = "Any",
    output_type: str = "Any",
    batch: bool = False,
    version: Optional[str] = None,
    meta_file: Optional[str] = None,
    args_schema: Optional[str] = None,
    args_schema_version: Optional[str] = None,
) -> BaseInterface:
    optionals = {}
    if version:
        optionals["version"] = version
    if meta_file:
        with open(meta_file, "r") as f:
            optionals["meta"] = json.load(f)
    if args_schema:
        with open(args_schema, "r") as f:
            optionals["args_schema"] = json.load(f)
    if args_schema_version:
        optionals["args_schema_version"] = args_schema_version

    payload = TaskCreate(
        name=name,
        command=command,
        source=source,
        input_type=input_type,
        output_type=output_type,
        **optionals,
    ).dict(exclude_unset=True)
    res = await client.post(f"{settings.BASE_URL}/task/", json=payload)
    new_task = check_response(res, expected_status_code=201, coerce=TaskRead)

    if batch:
        return PrintInterface(retcode=0, data=str(new_task.id))
    else:
        return RichJsonInterface(retcode=0, data=new_task.dict())


async def patch_task(
    client: AuthClient,
    *,
    id: Optional[int] = None,
    name: Optional[str] = None,
    version: Optional[str] = None,
    new_name: Optional[str] = None,
    new_command: Optional[str] = None,
    new_input_type: Optional[str] = None,
    new_output_type: Optional[str] = None,
    new_version: Optional[str] = None,
    meta_file: Optional[str] = None,
    new_args_schema: Optional[str] = None,
    new_args_schema_version: Optional[str] = None,
) -> BaseInterface:

    if id:
        if version:
            logging.error(
                "Too many arguments: cannot provide both `id` and `version`."
            )
            sys.exit(1)
    else:
        try:
            id = await get_task_id_from_cache(
                client=client, task_name=name, version=version
            )
        except FractalCacheError as e:
            print(e)
            sys.exit(1)

    update = {}
    if new_name:
        update["name"] = new_name
    if new_input_type:
        update["input_type"] = new_input_type
    if new_output_type:
        update["output_type"] = new_output_type
    if new_command:
        update["command"] = new_command
    if new_version:
        update["version"] = new_version
    if meta_file:
        with open(meta_file, "r") as f:
            update["meta"] = json.load(f)
    if new_args_schema:
        with open(new_args_schema, "r") as f:
            update["args_schema"] = json.load(f)
    if new_args_schema_version:
        update["args_schema_version"] = new_args_schema_version

    task_update = TaskUpdate(**update)
    payload = task_update.dict(exclude_unset=True)
    if not payload:
        return PrintInterface(retcode=1, data="Nothing to update")

    res = await client.patch(f"{settings.BASE_URL}/task/{id}", json=payload)
    new_task = check_response(res, expected_status_code=200, coerce=TaskRead)
    return RichJsonInterface(retcode=0, data=new_task.dict())


async def delete_task(
    client: AuthClient,
    *,
    id: Optional[int] = None,
    name: Optional[str] = None,
    version: Optional[str] = None,
) -> PrintInterface:

    if id:
        if version:
            logging.error(
                "Too many arguments: cannot provide both `id` and `version`."
            )
            sys.exit(1)
    else:
        try:
            id = await get_task_id_from_cache(
                client=client, task_name=name, version=version
            )
        except FractalCacheError as e:
            print(e)
            sys.exit(1)
    res = await client.delete(f"{settings.BASE_URL}/task/{id}")
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")



import json
import os
from typing import Optional

from rich.table import Table

from ..authclient import AuthClient
from ..common.schemas import DatasetCreate
from ..common.schemas import DatasetRead
from ..common.schemas import DatasetUpdate
from ..common.schemas import ResourceCreate
from ..common.schemas import WorkflowExport
from ..common.schemas import ResourceRead
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichConsoleInterface
from ..interface import RichJsonInterface
from ..response import check_response


async def post_dataset(
    client: AuthClient,
    *,
    project_id: int,
    dataset_name: str,
    metadata: Optional[str] = None,
    type: Optional[str] = None,
    batch: bool = False,
) -> RichJsonInterface:
    """
    Arguments:
        project_id: ID of project to add the new dataset to
        dataset_name: Name of new dataset
        metadata: Path to file containing dataset metadata in JSON format.
        type: Dataset type.
    """
    if metadata is None:
        meta = {}
    else:
        with open(metadata, "r") as f:
            meta = json.load(f)

    dataset_dict = dict(name=dataset_name, meta=meta)
    if type:
        dataset_dict["type"] = type
    dataset = DatasetCreate(**dataset_dict)

    res = await client.post(
        f"{settings.BASE_URL}/project/{project_id}/dataset/",
        json=dataset.dict(exclude_unset=True),
    )
    new_dataset = check_response(
        res, expected_status_code=201, coerce=DatasetRead
    )
    if batch:
        return PrintInterface(retcode=0, data=new_dataset.id)
    else:
        return RichJsonInterface(retcode=0, data=new_dataset.dict())


async def post_resource(
    client: AuthClient,
    *,
    project_id: int,
    dataset_id: int,
    path: str,
    batch: bool = False,
) -> BaseInterface:

    # Check that path is absolute, which is needed for when the server submits
    # tasks as a different user
    if not os.path.isabs(path):
        msg = f"{path=} is not an absolute path"
        raise ValueError(msg)

    resource = ResourceCreate(path=path)

    res = await client.post(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"dataset/{dataset_id}/resource/"
        ),
        json=resource.dict(),
    )
    new_resource = check_response(
        res, expected_status_code=201, coerce=ResourceRead
    )
    if batch:
        return PrintInterface(retcode=0, data=new_resource.id)
    else:
        return RichJsonInterface(retcode=0, data=new_resource.dict())


async def delete_resource(
    client: AuthClient,
    *,
    project_id: int,
    dataset_id: int,
    resource_id: int,
) -> BaseInterface:
    res = await client.delete(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"dataset/{dataset_id}/resource/{resource_id}"
        )
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


async def patch_dataset(
    client: AuthClient,
    *,
    project_id: int,
    dataset_id: int,
    new_name: Optional[str] = None,
    new_type: Optional[str] = None,
    meta_file: Optional[str] = None,
    make_read_only: bool = False,
    remove_read_only: bool = False,
) -> BaseInterface:

    # Prepare payload
    dataset_update_dict = {}
    if new_name:
        dataset_update_dict["name"] = new_name
    if new_type:
        dataset_update_dict["type"] = new_type
    if meta_file:
        with open(meta_file, "r") as f:
            meta = json.load(f)
        dataset_update_dict.update(meta=meta)
    if make_read_only:
        dataset_update_dict["read_only"] = True
    if remove_read_only:
        dataset_update_dict["read_only"] = False
    dataset_update = DatasetUpdate(**dataset_update_dict)
    payload = dataset_update.dict(exclude_unset=True)

    if not payload:
        return PrintInterface(retcode=1, data="Nothing to update")

    res = await client.patch(
        (f"{settings.BASE_URL}/project/{project_id}/" f"dataset/{dataset_id}"),
        json=payload,
    )
    new_dataset = check_response(
        res, expected_status_code=200, coerce=DatasetRead
    )
    return RichJsonInterface(retcode=0, data=new_dataset.dict())


async def get_dataset(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> BaseInterface:
    res = await client.get(
        f"{settings.BASE_URL}/project/{project_id}/dataset/{dataset_id}"
    )
    from rich.console import Group

    dataset = check_response(res, expected_status_code=200, coerce=DatasetRead)

    table = Table(title="Dataset")
    table.add_column("ID", style="cyan", no_wrap=True)
    table.add_column("Name", justify="right", style="green")
    table.add_column("Type", style="white")
    table.add_column("Meta", justify="center")
    table.add_column("Read only", justify="center")

    table.add_row(
        str(dataset.id),
        dataset.name,
        dataset.type,
        json.dumps(dataset.meta, indent=2),
        "✅" if dataset.read_only else "❌",
    )
    table_res = Table(title="Resources")
    table_res.add_column("Path", justify="center", style="yellow")
    table_res.add_column("ID", justify="center", style="yellow")
    table_res.add_column("Dataset ID", justify="center", style="yellow")
    for r in dataset.resource_list:
        table_res.add_row(r.path, str(r.id), str(r.dataset_id))
    group = Group(table, table_res)
    return RichConsoleInterface(retcode=0, data=group)


async def delete_dataset(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> PrintInterface:

    res = await client.delete(
        f"{settings.BASE_URL}/project/{project_id}/dataset/{dataset_id}"
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


async def get_dataset_history(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> BaseInterface:
    res = await client.get(
        f"{settings.BASE_URL}/project/{project_id}/dataset/{dataset_id}/"
        "export_history/"
    )

    history_workflow = check_response(
            res,
            expected_status_code=200,
            coerce=WorkflowExport,
            )
    return RichJsonInterface(retcode=0, data=history_workflow.dict())


async def get_dataset_status(
    client: AuthClient, *, project_id: int, dataset_id: int
) -> BaseInterface:
    res = await client.get(
        f"{settings.BASE_URL}/project/{project_id}/dataset/{dataset_id}/"
        "status/"
    )

    dataset_status = check_response(
            res,
            expected_status_code=200,
            )
    return RichJsonInterface(retcode=0, data=dataset_status)



from typing import Any
from typing import Optional

from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from ._validators import val_absolute_path
from ._validators import valstr


__all__ = (
    "ProjectCreate",
    "ProjectRead",
    "ProjectUpdate",
    "DatasetUpdate",
    "DatasetCreate",
    "DatasetRead",
    "ResourceCreate",
    "ResourceRead",
    "ResourceUpdate",
)


# RESOURCE


class _ResourceBase(BaseModel):
    """
    Base class for Resource
    """

    path: str


class ResourceCreate(_ResourceBase):
    # Validators
    _path = validator("path", allow_reuse=True)(val_absolute_path("path"))


class ResourceUpdate(_ResourceBase):
    # Validators
    _path = validator("path", allow_reuse=True)(val_absolute_path("path"))


class ResourceRead(_ResourceBase):
    id: int
    dataset_id: int


# DATASET


class _DatasetBase(BaseModel):
    """
    Base class for Dataset

    Attributes:
        name: TBD
        type: TBD
        meta: TBD
        read_only: TBD
    """

    name: str
    type: Optional[str]
    meta: dict[str, Any] = Field(default={})
    read_only: bool = False


class DatasetUpdate(_DatasetBase):
    name: Optional[str]
    meta: Optional[dict[str, Any]] = None
    read_only: Optional[bool]

    # Validators
    _name = validator("name", allow_reuse=True)(valstr("name"))
    _type = validator("type", allow_reuse=True)(valstr("type"))


class DatasetCreate(_DatasetBase):
    # Validators
    _name = validator("name", allow_reuse=True)(valstr("name"))
    _type = validator("type", allow_reuse=True)(valstr("type"))


class DatasetRead(_DatasetBase):
    id: int
    resource_list: list[ResourceRead]
    project_id: int
    read_only: bool


# PROJECT


class _ProjectBase(BaseModel):
    """
    Base class for Project

    Attributes:
        name: TBD
        read_only: TBD
    """

    name: str
    read_only: bool = False


class ProjectCreate(_ProjectBase):
    default_dataset_name: Optional[str] = "default"

    # Validators
    _name = validator("name", allow_reuse=True)(valstr("name"))
    _default_dataset_name = validator(
        "default_dataset_name", allow_reuse=True
    )(valstr("default_dataset_name"))


class ProjectRead(_ProjectBase):
    id: int
    dataset_list: list[DatasetRead] = []


class ProjectUpdate(_ProjectBase):
    name: Optional[str]
    read_only: Optional[bool]

    # Validators
    _name = validator("name", allow_reuse=True)(valstr("name"))
